/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne (dave@aifh.ed.ac.uk),                   *
 *           updated by  Emma Collingwood (emmac@aisb.ed.ac.uk)         *
 *           built in PGA, which was written by Geoffrey H. Ballinger   *
 *             (geoff@ed.ac.uk), and Peter Ross (peter@ed.ac.uk)        *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Geoffrey H. Ballinger.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

/*                                                                        *
 *  Filename: cross.c                                                     * 
 *                                                                        *
 * Contains functions required to perform various types ofcrossover on    *
 * two chromosomes, to produce a single child chromosome:                 *
 *                                                                        *
 *  uniform crossover                                                     *
 *  two-point crossover                                                   *
 *  one-point crossover                                                   *
 *  temporal_stamp                                                        *
 *  directed_temporal_stamp_recombination                                 *
 *  violation_directed_recombination                                      *
 *
*/


#include "gatt.h"
#include <string.h>

extern double sqrt ();
extern int geno_size;
extern int chromo_id;
extern int twins;
extern int roomsmatter;
extern int agentsmatter;
extern long random ();
extern char *malloc ();
extern FILE *output;
extern int nummods;
extern int n_formae;
extern int n_vdr;
extern int numjuxts;
extern int numedges;
extern int slotseatconstraints;
extern int sloteventconstraints;
extern double drandom();

/************************************************************************/
/* This performs one point crossover on 'parent1' and 'parent2'.        */
/* Child is new (even though there may be identical genotypes elsewhere */
/* in the system) so has refcount 0; set to 1 only upon insertion       */
/************************************************************************/

CHROMOPAIR
one_pt_cross (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  int chosen, i;
  CHROMOPAIR children;

  chosen = random ()% geno_size;/* Generate crossover point randomly. */


  /* Create the child. */
  children.child1.gdp = (GENODATA *) malloc (sizeof (GENODATA));

  if (children.child1.gdp == NULL)
    {
      printf ("Not enough memory\n");
      exit (0);
    }

  children.child1.gdp->refcount = 0;
  children.child1.gdp->genotype = (char *) malloc (geno_size + 1);
  children.child1.gdp->genotype[geno_size] = '\0';
  children.child1.gdp->id = chromo_id++;
  children.child1.gdp->parent1 = parent1.gdp->id;
  children.child1.gdp->parent2 = parent2.gdp->id;

  if (twins)
    {
      children.child2.gdp = (GENODATA *) malloc (sizeof (GENODATA));
      children.child2.gdp->refcount = 0;
      children.child2.gdp->genotype = (char *) malloc (geno_size + 1);
      children.child2.gdp->genotype[geno_size] = '\0';
      children.child2.gdp->id = chromo_id++;
      children.child2.gdp->parent1 = parent1.gdp->id;
      children.child2.gdp->parent2 = parent2.gdp->id;
    }
  /* Do the crossover */
  for (i = 0; i < chosen; i++)
    {
      children.child1.gdp->genotype[i] = parent2.gdp->genotype[i];
      if (twins)
	children.child2.gdp->genotype[i] = parent1.gdp->genotype[i];
    }
  for (i = chosen; i < geno_size; i++)
    {
      children.child1.gdp->genotype[i] = parent1.gdp->genotype[i];
      if (twins)
	children.child2.gdp->genotype[i] = parent2.gdp->genotype[i];
    }
  return (children);
}

/************************************************************************/
/* This performs two point crossover on 'parent1' and 'parent2'.        */
/* Child is new (even though there may be identical genotypes elsewhere */
/* in the system) so has refcount 0; set to 1 only upon insertion       */
/************************************************************************/

CHROMOPAIR
two_pt_cross (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  int first, last, i, swap;
  CHROMOPAIR children;

  first = random ()% geno_size;	/* Generate the two points randomly. */
  last = random ()% geno_size;

  if (first > last)
    {				/* Make sure that they are in the */
      swap = first;		/* right order.                   */
      first = last;
      last = swap;
    }

  /* Create the child. */
  children.child1.gdp = (GENODATA *) malloc (sizeof (GENODATA));
  if (children.child1.gdp == (GENODATA *) NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  children.child1.gdp->refcount = 0;
  children.child1.gdp->genotype = (char *) malloc (geno_size + 1);
  if (children.child1.gdp->genotype == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  children.child1.gdp->genotype[geno_size] = '\0';
  children.child1.gdp->id = chromo_id++;
  children.child1.gdp->parent1 = parent1.gdp->id;
  children.child1.gdp->parent2 = parent2.gdp->id;

  if (twins)
    {
      children.child2.gdp = (GENODATA *) malloc (sizeof (GENODATA));
      if (children.child2.gdp == (GENODATA *) NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}
      children.child2.gdp->refcount = 0;
      children.child2.gdp->genotype = (char *) malloc (geno_size + 1);
      if (children.child2.gdp->genotype == NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}


      children.child2.gdp->genotype[geno_size] = '\0';
      children.child2.gdp->id = chromo_id++;
      children.child2.gdp->parent1 = parent1.gdp->id;
      children.child2.gdp->parent2 = parent2.gdp->id;
    }

  /* Do the crossover */
  for (i = 0; i < first; i++)
    {
      children.child1.gdp->genotype[i] = parent2.gdp->genotype[i];
      if (twins)
	children.child2.gdp->genotype[i] = parent1.gdp->genotype[i];
    }
  for (i = first; i < last; i++)
    {
      children.child1.gdp->genotype[i] = parent1.gdp->genotype[i];
      if (twins)
	children.child2.gdp->genotype[i] = parent2.gdp->genotype[i];
    }
  for (i = last; i < geno_size; i++)
    {
      children.child1.gdp->genotype[i] = parent2.gdp->genotype[i];
      if (twins)
	children.child2.gdp->genotype[i] = parent1.gdp->genotype[i];
    }

  return (children);
}

/************************************************************************/
/* This performs uniform crossover on 'parent1' and 'parent2'.          */
/* Uses rand_bit() to cut down on expensive random number generation.   */
/* Child is new (even though there may be identical genotypes elsewhere */
/* in the system) so has refcount 0; set to 1 only upon insertion       */
/************************************************************************/

int rand_bit_src;
int rand_bit_bits = 0;
int
rand_bit ()
{
  if (rand_bit_bits == 0)
    {
      rand_bit_bits = 16;
      rand_bit_src = random ()% 131072;	/* Get 16+1 more random bits */
    }
  rand_bit_src = rand_bit_src >> 1;	/* Lose 1 bit */
  rand_bit_bits--;		/* Number left after this call */
  return (rand_bit_src & 1);
}

CHROMOSOME
violation_directed_recombination (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  register int i, choice, worst, nchosen;
  int *chosen, tsize;
  char *alreadychosen;
  CHROMOSOME child;

  nschanges = nrchanges = nachanges = nchosen = 0;

  child.gdp = (GENODATA *) malloc (sizeof (GENODATA));
  if (child.gdp == (GENODATA *) NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp->genotype = (char *) malloc (geno_size + 1);
  if (child.gdp->genotype == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp->genotype[geno_size] = '\0';
  child.gdp->id = chromo_id++;
  child.gdp->parent1 = parent1.gdp->id;
  child.gdp->parent2 = parent2.gdp->id;

  if (roomsmatter)
    child.sloads = (int *) calloc (numrooms * numslots, sizeof (int));
  if (roomsmatter)
    child.eloads = (int *) calloc (numrooms * numslots, sizeof (int));

  if (roomsmatter)
    if (child.sloads == NULL || child.eloads == NULL)
      {
	printf ("Not enough memory\n");
	exit (1);
      }

  strncpy (child.gdp->genotype, parent1.gdp->genotype, geno_size + 1);
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;

  chosen = (int *) malloc (n_vdr * sizeof (int));
  if (chosen == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  alreadychosen = (char *) calloc (n_vdr, sizeof (char));
  if (alreadychosen == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  /* In violation directed recombination, we just want to find n_vdr genes */
  /* (without replacement) with fairly high violation scores in parent1,   */
  /* just give them the child the slots these have in parent2.             */

  /* First, pick n_vdr different genes with high violation scores, using a */
  /* tournament size of 2*sqrt(nummods) */

  tsize = 2 * (int) sqrt (nummods);


  while (nchosen < n_vdr)
    {
      choice = sunspecs[random ()% nsunspecs];	/* pre-specified slot */
      while (alreadychosen[choice])
	choice = sunspecs[random ()% nsunspecs];
      worst = choice;		/* This is worst so far */
      for (i = 1; i < tsize; i++)
	{
	  choice = sunspecs[random ()% nsunspecs];
	}

      chosen[nchosen++] = worst;
      alreadychosen[worst] = 1;
    }

  /* n_vdr genes have been chosen to have their places switched */
  /* So now do the switch */

  for (i = 0; i < n_vdr; i++)
    child.gdp->genotype[chosen[i]] = parent2.gdp->genotype[chosen[i]];

  for (i = 0; i < nummods; i++)
    if (parent1.gdp->genotype[i] != child.gdp->genotype[i])
      {
	schanges[nschanges++] = i;
	schanged[i] = 1;
      }

  free (chosen);
  free (alreadychosen);
  return (child);
}


CHROMOSOME
directed_temporal_stamp_recombination (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  register int i, e1, e2, s1, s2, ss1, ss2, sd, constraint_choice, transmit, fix_e1,
    nsschanges, choice, worst;
  int *sschanges;
  CHROMOSOME child;
  _MODULE *m;
  LL_EDGES *edgechosen;
  LL_JUXTS *juxtchosen;

  nschanges = nrchanges = nachanges = nsschanges = 0;

  sschanges = (int *) malloc (nummods * sizeof (int));
  if (sschanges == (int *) NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp = (GENODATA *) malloc (sizeof (GENODATA));
  if (child.gdp == (GENODATA *) NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp->genotype = (char *) malloc (geno_size + 1);
  if (child.gdp->genotype == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp->genotype[geno_size] = '\0';
  child.gdp->id = chromo_id++;
  child.gdp->parent1 = parent1.gdp->id;
  child.gdp->parent2 = parent2.gdp->id;

  if (roomsmatter)
    child.sloads = (int *) calloc (numrooms * numslots, sizeof (int));
  if (roomsmatter)
    child.eloads = (int *) calloc (numrooms * numslots, sizeof (int));

  if (roomsmatter)
    if (child.sloads == NULL || child.eloads == NULL)
      {
	printf ("Not enough memory\n");
	exit (1);
      }


  if (slotseatconstraints){
    child.ssloads = (int *) calloc (numslots, sizeof (int));
    if (child.ssloads == NULL){
      printf ("Not enough memory\n");
	exit (1);
      }
  }

  if (sloteventconstraints){
    child.seloads = (int *) calloc (numslots, sizeof (int));
    if (child.seloads == NULL) {
	printf ("Not enough memory\n");
	exit (1);
      }
  }
  strncpy (child.gdp->genotype, parent1.gdp->genotype, geno_size + 1);
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;

  transmit = 0;

  while (transmit < n_formae)
    {
      /* First, choose a well-placed gene  */
      worst = random ()% nummods;

      choice = random ()% nummods;

      m = a_modules[worst];

      /* Now randomly choose one of its constraints */

      constraint_choice = random ()% (m->numedges + m->numjuxts);
      if (constraint_choice < m->numjuxts)
	{
	  i = 0;
	  juxtchosen = m->juxts;
	  while (i++ < constraint_choice)
	    juxtchosen = juxtchosen->next;
	  e1 = juxtchosen->j->a->id;
	  e2 = juxtchosen->j->b->id;
	}
      else
	{
	  i = 0;
	  edgechosen = m->edges;
	  while (i++ < (constraint_choice - m->numjuxts))
	    edgechosen = edgechosen->next;
	  e1 = edgechosen->e->mod1->id;
	  e2 = edgechosen->e->mod2->id;
	}
      s1 = parent1.gdp->genotype[e1];
      s2 = parent1.gdp->genotype[e2];
      ss1 = parent2.gdp->genotype[e1];
      ss2 = parent2.gdp->genotype[e2];

      sd = ss2 - ss1;

      if (drandom ()< 0.5)
	fix_e1 = 1;
      else
	fix_e1 = 0;

      if (fix_e1)
	{			/* Stamp temporal difference with e1 fixed */
	  if (((s1 + sd) > -1) && (s1 + sd) < numslots)
	    {			/* If appropriate */
	      child.gdp->genotype[e2] = s1 + sd;
	      if (parent1.gdp->genotype[e2] != (s1 + sd))
		{
		  sschanges[nsschanges++] = e2;
		}
	    }
	  else if (((s2 - sd) > -1) && (s2 - sd) < numslots)
	    {			/* fix e2 instead */
	      child.gdp->genotype[e1] = s2 - sd;
	      if (parent1.gdp->genotype[e1] != (s2 - sd))
		{
		  sschanges[nsschanges++] = e1;
		}
	    }
	}
      else
	{
	  if (((s2 - sd) > -1) && (s2 - sd) < numslots)
	    {			/* If e2 fixable   */
	      child.gdp->genotype[e1] = s2 - sd;
	      if (parent1.gdp->genotype[e1] != (s2 - sd))
		{
		  sschanges[nsschanges++] = e1;
		}
	    }
	  else if (((s1 + sd) > -1) && (s1 + sd) < numslots)
	    {			/* fix e1 instead */
	      child.gdp->genotype[e2] = s1 + sd;
	      if (parent1.gdp->genotype[e2] != (s1 + sd))
		{
		  sschanges[nsschanges++] = e2;
		}
	    }
	}
      transmit++;
    }

  /* Now find out which are the real changes, and sort them out */

  for (i = 0; i < nummods; i++)
    if (parent1.gdp->genotype[i] != child.gdp->genotype[i])
      {
	schanges[nschanges++] = i;
	schanged[i] = 1;
      }

  free (sschanges);
  return (child);
}


CHROMOSOME
temporal_stamp_recombination (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  register int i, e1, e2, s1, s2, ss1, ss2, sd, constraint_choice, transmit, fix_e1,
    nsschanges;
  int *sschanges;
  CHROMOSOME child;

  nschanges = nrchanges = nachanges = nsschanges = 0;

  sschanges = (int *) malloc (nummods * sizeof (int));
  if (sschanges == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp = (GENODATA *) malloc (sizeof (GENODATA));
  if (child.gdp == (GENODATA *) NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp->genotype = (char *) malloc (geno_size + 1);
  if (child.gdp->genotype == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  child.gdp->genotype[geno_size] = '\0';
  child.gdp->id = chromo_id++;
  child.gdp->parent1 = parent1.gdp->id;
  child.gdp->parent2 = parent2.gdp->id;

  if (roomsmatter)
    child.sloads = (int *) calloc (numrooms * numslots, sizeof (int));
  if (roomsmatter)
    child.eloads = (int *) calloc (numrooms * numslots, sizeof (int));

  if (roomsmatter)
    if (child.sloads == NULL || child.eloads == NULL)
      {
	printf ("Not enough memory\n");
	exit (1);
      }


  strncpy (child.gdp->genotype, parent1.gdp->genotype, geno_size + 1);
  for (i = 0; i < nummods; i++)
    schanged[i] = 0;

  transmit = 0;

  /* In temporal stamp recombination (TSR) we do the following, using  */
  /* the parameter n_formae (number of formae to use):                 */
  /*     Repeat n_formae times                                         */
  /*       Let c be a randomly chosen binary constraint, with          */
  /*         events e1 and e2. Slots in parent1 are s1 and s2, slots   */
  /*         in parent2 are ss1 and ss2.                               */
  /*       Let sd be s2-s1.                                            */
  /*       Randomly choose e1 or e2                                    */
  /*          If e1 chosen, let ss1 in the child be unchanged, but     */
  /*            ss2 becomes appropriate_slot(ss1+sd)                   */
  /*          If e2 chosen, let ss2 in child be unchanged, but ss1     */
  /*            becomes appropriate_slot(ss2-sd)                       */
  /*       If the chosen event now has a different slot from in parent1 */
  /*        then note this fact.                                        */
  /*                                                                   */
  /*   appropriate_slot(S) is S if S is a valid slot (0<=S<=numslots)  */
  /*   If this doesn't work, choose the other event and try again. If  */
  /*   still a problem, forget it                                      */


  while (transmit < n_formae)
    {
      constraint_choice = random ()% (numjuxts + numedges);
      if (constraint_choice < numjuxts)
	{
	  e1 = a_juxts[constraint_choice]->a->id;
	  e2 = a_juxts[constraint_choice]->b->id;
	}
      else
	{
	  e1 = a_edges[constraint_choice - numjuxts]->mod1->id;
	  e2 = a_edges[constraint_choice - numjuxts]->mod2->id;
	}
      s1 = parent1.gdp->genotype[e1];
      s2 = parent1.gdp->genotype[e2];
      ss1 = parent2.gdp->genotype[e1];
      ss2 = parent2.gdp->genotype[e2];

      sd = ss2 - ss1;

      if (drandom ()< 0.5)
	fix_e1 = 1;
      else
	fix_e1 = 0;

      if (fix_e1)
	{			/* Stamp temporal difference with e1 fixed */
	  if (((s1 + sd) > -1) && (s1 + sd) < numslots)
	    {			/* If appropriate */
	      child.gdp->genotype[e2] = s1 + sd;
	      if (parent1.gdp->genotype[e2] != (s1 + sd))
		{
		  sschanges[nsschanges++] = e2;
		}
	    }
	  else if (((s2 - sd) > -1) && (s2 - sd) < numslots)
	    {			/* fix e2 instead */
	      child.gdp->genotype[e1] = s2 - sd;
	      if (parent1.gdp->genotype[e1] != (s2 - sd))
		{
		  sschanges[nsschanges++] = e1;
		}
	    }
	}
      else
	{
	  if (((s2 - sd) > -1) && (s2 - sd) < numslots)
	    {			/* If e2 fixable   */
	      child.gdp->genotype[e1] = s2 - sd;
	      if (parent1.gdp->genotype[e1] != (s2 - sd))
		{
		  sschanges[nsschanges++] = e1;
		}
	    }
	  else if (((s1 + sd) > -1) && (s1 + sd) < numslots)
	    {			/* fix e1 instead */
	      child.gdp->genotype[e2] = s1 + sd;
	      if (parent1.gdp->genotype[e2] != (s1 + sd))
		{
		  sschanges[nsschanges++] = e2;
		}
	    }
	}
      transmit++;
    }

  /* Now find out which are the real changes, and sort them out */

  for (i = 0; i < nummods; i++)
    if (parent1.gdp->genotype[i] != child.gdp->genotype[i])
      {
	schanges[nschanges++] = i;
	schanged[i] = 1;
      }

  free (sschanges);
  return (child);
}

CHROMOSOME
permute_slots (parent)
     CHROMOSOME parent;
{
  int i, j, *newslots, *taken, temp;
  CHROMOSOME child;

  nschanges = nrchanges = nachanges = 0;

  child.gdp = (GENODATA *) malloc (sizeof (GENODATA));
  if (child.gdp == (GENODATA *) NULL)
    {
      printf ("Not enough memory \n");
      exit (1);
    }


  child.gdp->genotype = (char *) malloc (geno_size + 1);
  if (child.gdp->genotype == NULL)
    {
      printf ("Not enough memory \n");
      exit (1);
    }

  child.gdp->genotype[geno_size] = '\0';
  child.gdp->id = chromo_id++;
  child.gdp->parent1 = parent.gdp->id;
  child.gdp->parent2 = parent.gdp->id;

  if (roomsmatter)
    {
      child.sloads = (int *) calloc (numrooms * numslots, sizeof (int));
      child.eloads = (int *) calloc (numrooms * numslots, sizeof (int));

      if (child.sloads == NULL || child.eloads == NULL)
	{
	  printf ("Not enough memory \n");
	  exit (1);
	}
    }

  if (slotseatconstraints){
    child.ssloads = (int *) calloc (numslots, sizeof (int));
    if (child.ssloads == NULL){
      printf ("Not enough memory \n");
      exit (1);
    }
  }

  if (sloteventconstraints){
    child.seloads = (int *) calloc (numslots, sizeof (int));
    if (child.seloads == NULL) {
      printf ("Not enough memory \n");
      exit (1);
    }
  }

  /* Now, make the child `same' as parent but with randomly permuted slots */

  /* Randomly decide which slot will change to which */

  newslots = (int *) malloc (numslots * sizeof (int));
  if (newslots == NULL)
    {
      printf ("Not enough memory \n");
      exit (1);
    }

  taken = (int *) malloc (numslots * sizeof (int));
  if (taken == NULL)
    {
      printf ("Not enough memory \n");
      exit (1);
    }

  for (i = 0; i < numslots; i++)
    newslots[i] = random ()% numslots;
  for (i = 0; i < numslots; i++)
    taken[i] = i;
  for (i = 0; i < numslots; i++)
    {
      temp = newslots[i] % (numslots - i);
      newslots[i] = taken[temp];
      for (j = temp; j < numslots - i - 1; j++)
	taken[j] = taken[j + 1];
    }

  for (i = 0; i < nummods; i++)
    {
      child.gdp->genotype[i] = newslots[parent.gdp->genotype[i]];
      if (parent.gdp->genotype[i] != child.gdp->genotype[i])
	{
	  schanges[nschanges++] = i;
	  schanged[i] = 1;
	}
      else
	schanged[i] = 0;
    }
  free (newslots);
  free (taken);
  return (child);
}

CHROMOSOME
uniform_cross (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  int i;
  CHROMOSOME child;

  nschanges = nrchanges = nachanges = 0;

  child.gdp = (GENODATA *) malloc (sizeof (GENODATA));
  child.gdp->genotype = (char *) malloc (geno_size + 1);
  child.gdp->genotype[geno_size] = '\0';
  child.gdp->id = chromo_id++;
  child.gdp->parent1 = parent1.gdp->id;
  child.gdp->parent2 = parent2.gdp->id;

  if (roomsmatter)
    {
      child.sloads = (int *) calloc (numrooms*numslots, sizeof (int));
      child.eloads = (int *) calloc (numrooms*numslots, sizeof (int));
      if (child.sloads == NULL || child.eloads == NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}
    }

  if (agentsmatter)
    {
      child.aeloads = (int *) calloc (numagents, sizeof (int));
      child.ahloads = (int *) calloc (numagents, sizeof (int));
      if (child.aeloads == NULL || child.ahloads == NULL)
	{
	  printf ("Not enough memory\n");
	  exit (1);
	}
    }
  

  if (slotseatconstraints){
    child.ssloads = (int *) calloc (numslots, sizeof (int));
    if (child.ssloads == NULL){
      printf ("Not enough memory\n");
      exit (1);
    }
  }

  if (sloteventconstraints){
    child.seloads = (int *) calloc (numslots, sizeof (int));
    if (child.seloads == NULL){
      printf ("Not enough memory\n");
      exit (1);
    }
  }

  /* Do the crossover */
  for (i = 0; i < nummods; i++)
    if (rand_bit ())
      {
	child.gdp->genotype[i] = parent2.gdp->genotype[i];
	if (parent2.gdp->genotype[i] != parent1.gdp->genotype[i])
	  {
	    schanges[nschanges++] = i;
	    schanged[i] = 1;
	  }
	else
	  schanged[i] = 0;
      }
    else
      {
	child.gdp->genotype[i] = parent1.gdp->genotype[i];
	schanged[i] = 0;
      }
  if (roomsmatter)
    for (i = nummods; i < 2 * nummods; i++)
      if (rand_bit ())
	{
	  child.gdp->genotype[i] = parent2.gdp->genotype[i];
	  if (parent2.gdp->genotype[i] != parent1.gdp->genotype[i])
	    {
	      rchanges[nrchanges++] = i - nummods;
	      rchanged[i - nummods] = 1;
	    }
	  else
	    rchanged[i - nummods] = 0;
	}
      else
	{
	  child.gdp->genotype[i] = parent1.gdp->genotype[i];
	  rchanged[i - nummods] = 0;
	}

  if (agentsmatter)
    for (i = nummods + (roomsmatter * nummods);
	 i < (nummods + (roomsmatter * nummods) + nummods); i++)
      if (rand_bit ())
	{
	  child.gdp->genotype[i] = parent2.gdp->genotype[i];
	  if (parent2.gdp->genotype[i] != parent1.gdp->genotype[i])
	    {
	      achanges[nachanges++] = (i - nummods) - (roomsmatter * nummods);
	      achanged[(i - nummods) - (roomsmatter * nummods)] = 1;
	    }
	  else
	    achanged[(i - nummods) - (roomsmatter * nummods)] = 0;
	}
      else
	{
	  child.gdp->genotype[i] = parent1.gdp->genotype[i];
	  achanged[(i - nummods) - (roomsmatter * nummods)] = 0;
	}
  return (child);
}
