/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


#include <stdio.h>            /* Include the IO and maths header files. */
#include <math.h>             /* Include maths stuff */
#include <curses.h>           /* screen output stuff */
#include "screen.h"           /* #defines for screen layout configuration */


#ifdef SOLARIS
#include <stdlib.h>
#define random() lrand48()
#define srandom(seed) srand48((long) seed)
#endif

#ifdef SOLARIS 
extern int sprintf();
#else
extern char *sprintf();
#endif SOLARIS

#define MAXPOPS 100
#define MAXNAME 20
#define MAXINNUMS 100
#define MAXDAYS 40
#define MAXSLOTS 120
#define MAXROOMS 50
#define MAXSTUDENTS 20000
#define MAXMODS 1000
#define MAXOPTIONS 20
#define MAXTAKES 20
#define MAXPRIORITY 100
#define MAXAGENTS 100
#define MAXINPUTNAMES 100
#define MAXPREF  5        /* Maximum value for agent preferences */
#define MAXSIMS 50
#define MAXMEMS 1500
#define MAXSIMGROUPS 100
#define MAXSUBPROBS  50
/*************************************************************************/
/* GA-specific structures ************************************************/
/*************************************************************************/

typedef struct genodata {     /* genotype string plus reference count    */
  char *genotype;
  char *slotperm;
  int refcount;
  int id;
  int parent1, parent2;
} GENODATA;

typedef struct chromosome {   /* A chromosome consists of a genotype and */
  GENODATA *gdp;               /* a fitness.                             */
  double fitness;
  int clashes, toocloses, modexcslots,           /* And a large collection  */
      modavslots, juxtviols, juxtscore,          /* of counters for various */
      eventcapviols, studcapviols, nperday,      /* types of constraint     */
      modexcrooms, modavrooms, agentexcslots,    /* violations.             */
    agentavslots, agentoverh, agentunderh, nclash_exams,
    agentovere, agentundere, agentexcmod, student_over,consecutives,
      agentavmod, agentprefs, agentclash,agenttoomany, ssload, seload;
  int *ssloads, *seloads;          /* Slot seat and event capacity loads     */
  int *sloads, *eloads;      /* Seat loads and event loads by room & slot */
    int *aeloads, *ahloads;    /* Teaching loads */
  int nclash, ntooclose;
} CHROMOSOME;

typedef struct chromopair {   /* Returned by crossover, contains one or */
  CHROMOSOME child1;          /* two children depending on -t flag      */
  CHROMOSOME child2;
} CHROMOPAIR;

typedef struct op_result {
  CHROMOSOME child;
  int newbest;
} OP_RESULT;

/* User actions possible */

#define QUIT      1
#define CONTINUE  2
#define RESTART   3
#define ASK_AGAIN 4
#define PRINT_TT  5

/* data input files */

int numslots;
int numdays;
int numrooms;
int nummods;
int numstudents;
int numagents;

int adayofday[MAXSLOTS];  /* The absolute day id of a day */


typedef struct spec {
       int forced;        /* Is this specification forced? */
       int id;            /* The slot, room, or whatever */
       int priority;      /* priority */
       struct spec *next;
     } LL_SPEC;


struct date {
       int day;
       int month;
       int year;
};

struct date *dates[MAXDAYS];
 
struct slot {
       int id, day, aday, time, sscap, secap;
     };
struct slot *slots[MAXSLOTS]; 

struct simgroup {
        
       int num, sims[MAXSIMS], size, total;

};
struct simgroup *simgroups[MAXSIMGROUPS];

struct subprob {
       int num, mems[MAXMEMS];
}; 

struct subprob *subprobs[MAXSUBPROBS];

struct cap {
       int slot; 
       int cap;
     };

struct room {
       char *name;
       int studcap;
       int modcap;
       int numsexceptions;
       struct cap *sexceptions[MAXSLOTS];
       int nummexceptions;
       struct cap *mexceptions[MAXSLOTS];
       int numexcslots;
       int excslots[MAXSLOTS];
     }; 
struct room *rooms[MAXROOMS];

typedef struct n { int n; } N;

typedef struct ll_studs {
        struct student *st;
        struct ll_studs *next;
      } LL_STUDS;

typedef struct juxt {
                 struct module *a, *b;
                 int rel; /* 0=before,1=after,2 = sametime,3= between */
                          /* 4 = sameday */
                 long index;
                 int id, forced, priority;
                 int dsm; /* 0 = days, 1 = minutes */
                 int elm; /* 0 = exactly, 1 = at least, 2 = at most */
                 int gap;
                 struct juxt *left, *right, *next;
               } JUXT;

JUXT *juxts, *s_juxts, *p_juxts, **a_juxts, *t_juxts;


typedef struct ll_juxts {
        struct juxt * j;
        struct ll_juxts *next;
      } LL_JUXTS; 
        
typedef struct ll_edges {
        struct edge * e;
        struct ll_edges *next;
      } LL_EDGES;

typedef struct module {
       char *name;
       int id, oid;
       int size, len;
       int slotspec, dayspec, roomspec, agentspec;
       char attendees;

       int numexcslot, numexcday, numexcroom, numexcagent;
       LL_SPEC *exslots, *s_exslots, *p_exslots;     /* The excluded slots */
       LL_SPEC *exrooms, *s_exrooms, *p_exrooms;     /* The excluded rooms */
       LL_SPEC *exagents, *s_exagents, *p_exagents;  /* The excluded agents */

       int numtakers;
       LL_STUDS *s_studs, *p_studs, *studs;

       int numedges, numoedges;
       LL_EDGES *edges, *s_edges, *p_edges;
       LL_EDGES *oedges, *s_oedges, *p_oedges;

       int numjuxts, numojuxts;
       LL_JUXTS *juxts, *s_juxts, *p_juxts; 
       LL_JUXTS *ojuxts, *s_ojuxts, *p_ojuxts;

      int sim; /* 0 if not in a sim group, 1 if sim-base, 2 if sim-member */
      int simgroup;  /* the group it's in */
       struct module *left, *right, *next;
     }  _MODULE;

_MODULE *modules, **a_modules, *s_modules, *p_modules, *t_modules, *f_module;


typedef struct ll_mods {
        _MODULE *mod;
        struct ll_mods *next;
      } LL_MODS;

/**************************************************************************/
/* NOTE:  Students start out as a linked list rather than a binary tree.  */
/* It is assumed that there is precisely one line of `student data' for   */
/* each student, so that each such line is guaranteed to be a new student.*/
/* This makes a lot of things simpler. If this wasn't assumed, then, for  */
/* example, we'd have to search to see if each student was already there, */
/* which would be useless in the binary tree since they are likely to be  */
/* in alphabetical order anyway! .........................................*/
/**************************************************************************/

typedef struct student {
       char *name;
       int id;
       int numtaken;
       LL_MODS *s_mods, *p_mods, *mods;
       struct student *next;     
     } _STUDENT;

       /**************************************************/
       /* root,    start,        placeholder,    array   */
       /**************************************************/
_STUDENT *students, *s_students, *p_students, **a_students;   

struct option {
       int mod;
       int des;
     };

struct agent {
       char *name;
       int nummustdo;
       int mustdolist[MAXOPTIONS];
       int numcando;
       struct option *cando[MAXOPTIONS];
       int numexcslots;
       int excslots[MAXSLOTS];
       int numavslots;
       struct option *avslots[MAXSLOTS];
       int maxmods, minmods, maxhours, minhours;
       int maxpref;
  
};

struct agent *agents[MAXAGENTS];

typedef struct edge {
                int id; 
                struct module *mod1;
                struct module *mod2;
                int ns;             /* The number sharing this edge */
                struct edge *left, *right, *next;
              } EDGE ;

typedef struct simedge {
          short int group;
          struct module *mod2;
          int ns;     /* The number sharing this edge */
} SIMEDGE ;

EDGE *edges, *p_edges, *s_edges, *t_edges, *f_edges, **a_edges, *cce;


enum {
       _SLOT_,                    /* These indicate different chromosome */
       _SLOT_ROOM_,               /* formats, according to whether or not */
       _SLOT_TUTOR_,              /* we are assigning rooms and/or tutors */
       _SLOT_ROOM_TUTOR_,         /* as well as slots */
     };


int roomcaps[MAXROOMS][MAXSLOTS][2];

struct inputname {
        char *n ;
      };
struct inputname *inputnamelist[MAXINPUTNAMES];

typedef struct pen_record {
     int id, pen;
   } PEN_RECORD;

PEN_RECORD **spens, **rpens, **apens;

int *sunspecs, *runspecs, *aunspecs, nsunspecs, nrunspecs, naunspecs;

int *schanges, *rchanges, *achanges, nschanges, nrchanges, nachanges;
int *hotslots, *hotrooms, *hotagents, *eslots, *erooms, *eagents;
char *schanged, *rchanged, *achanged;

int *roomeloads, *roomsloads;

int *slot_alleles, *room_alleles, *agent_alleles;

int allmodsexclude[MAXSLOTS];



struct cce {unsigned short a,b;};
struct cce *cces[2000];

