/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


/***********************************************************************
 * Filename: mutate.c
 *
 * Functions to perform mutation on a given chromosome, returning
 * the new chromosome.
 *
 */

#include "gatt.h"

extern int geno_size;
extern int chromo_id;
extern double mute_rate;
extern double drandom ();

extern char *malloc ();
extern long random ();
extern char *strncpy ();

double diversity_measure;

/************************************************************************/
/* This flips a random bit in 'original'.                               */
/* If given chromo has non-zero refcount, then work on copy instead.    */
/* Refcounts get adjusted upon insertion, later.                        */
/************************************************************************/

CHROMOSOME 
mutate_binary (original, rate)
     CHROMOSOME original;
     double rate;
{
  int place;
  CHROMOSOME newchromo;

  if (original.gdp->refcount > 0)
    {
      newchromo.gdp = (GENODATA *) malloc (sizeof (GENODATA));
      newchromo.gdp->refcount = 0;
      newchromo.gdp->genotype = (char *) malloc (geno_size + 1);
      newchromo.gdp->genotype[geno_size] = '\0';
      newchromo.gdp->id = chromo_id++;
      newchromo.gdp->parent1 = original.gdp->parent1;
      newchromo.gdp->parent2 = original.gdp->parent2;
      strncpy (newchromo.gdp->genotype, original.gdp->genotype, geno_size);
    }
  else
    newchromo = original;

  for (place = 0; place < geno_size; place++)
    {
      if (drandom ()<= rate)
	if (newchromo.gdp->genotype[place] == 0)
	  newchromo.gdp->genotype[place] = 1;
	else
	  newchromo.gdp->genotype[place] = 0;
    }

  return (newchromo);
}

/************************************************************************/
/* Function mrate() points at one of the next two functions, depending  */
/* on whether adaptive is chosen or note. Adaptive mutation scales the  */
/* probability of mutating each bit by the degree of similarity of the  */
/* parents.                                                             */
/************************************************************************/

/************************************************************************/
/* This function returns the bit mutation rate appropriate to the given */
/* parents - a number in the range 0..mute_rate, depending on degree of */
/* similarity of parents.                                               */
/************************************************************************/

double 
adaptive (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  int loop, count;
  double prob;

  count = 0;
  for (loop = 0; loop < geno_size; loop += 1)
    {
      if (parent1.gdp->genotype[loop] == parent2.gdp->genotype[loop])
	count++;
    }
  /* Proportion = observed / total. */
  prob = ((double) count) / ((double) geno_size);

  return (prob * mute_rate);
}

double 
diversity (pops, base, range, sample)
     CHROMOSOME *pops;
     int base, range, sample;
{
  int i, j, *choices, *sum, count = 0;
  double prob;

  choices = (int *) malloc (sample * sizeof (int));	/* Get space */
  if (choices == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }

  sum = (int *) malloc (geno_size * sizeof (int));
  if (sum == NULL)
    {
      printf ("Not enough memory\n");
      exit (1);
    }
  for (i = 0; i < sample; i++)
    choices[i] = random ()% range;	/* Choose sample */
  for (i = 0; i < geno_size; i++)
    sum[i] = 0;

  for (i = 0; i < sample; i++)	/* Measure diversity */
    for (j = 0; j < geno_size; j++)	/* by comparing mean */
      sum[j] += pops[base + choices[i]].gdp->genotype[j];	/* chromo with one   */

  for (i = 0; i < geno_size; i++)
    if (sum[i] == sample * pops[base + choices[0]].gdp->genotype[i])
      count++;

  /* Proportion = observed / total. */
  prob = ((double) count) / ((double) geno_size);

  return (sqrt (prob));
}



/************************************************************************/
/* This function simply returns the fixed mutation rate.                */
/************************************************************************/

double 
nonadaptive (parent1, parent2)
     CHROMOSOME parent1, parent2;
{
  return (mute_rate);
}
