/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/



/**********************************************************************
 *  Filename: reinit.c
 *
 *  Functions to perform  re-initialisation on a population
 **********************************************************************/ 

#include "gatt.h"

extern int geno_size;
extern double reinit_rate;
extern double *total_abs_fitness;
extern double drandom ();
extern int compchromo ();
extern int *bestinpop;
extern void reporton ();
extern int numchromo;
extern int spatialchosen;
extern int hctrials;
extern void freechromo ();
extern void (*init_genotype) ();
extern CHROMOSOME (*mutate) ();
extern CHROMOSOME permute_slots ();
extern double (*evaluate) ();
void no_reinit ();
void converge_reinit ();
extern void hillclimb ();
int reinits = 0;


void no_reinit (pops, base, range, rrate)
     CHROMOSOME *pops;
     int base, range;
     double rrate;
{
  return;
}

/* COMPLETELY reinitialise a population */

void complete_reinit (pops, base, range, rrate)
     CHROMOSOME *pops;
     int base, range;
     double rrate;
{
  int chromo, pop, noreinit = 0;
  CHROMOSOME child, old;
  pop = (int) (base / range);
  total_abs_fitness[pop] = 0.0;

  for (chromo = 1; chromo < range; chromo++)
    {
      if (!(pops[base].fitness == pops[base + chromo].fitness))
	{
	  noreinit = 1;
	  break;
	}
    }

  if (noreinit)
    return;			/* If there are some differences, don't reinit */
  if (pops[base].fitness > 0.9)	/* if no differences, but all perfect, */
    return;			/* don't reinit */

  reinits++;

  /* Then reinitialise the population by saving the best, and making the */
  /* rest rrate% mutants of the best */

  for (chromo = base; chromo < base + range; chromo++)
    {
      old = pops[chromo];
      child = mutate (pops[chromo], rrate);
      child.fitness = evaluate (child.gdp->genotype,
				old.gdp->genotype,
				&child,
				&old,
				base / numchromo);
      freechromo (&pops[chromo]);
      pops[chromo] = child;
    }


  if (!spatialchosen)
    {
      qsort ((char *) ((int) pops + (base * sizeof (CHROMOSOME))),
	     range, sizeof (CHROMOSOME), compchromo);
      bestinpop[pop] = 0;
    }
  else
    {
      bestinpop[pop] = 0;
      for (chromo = 0; chromo < numchromo; chromo++)
	{
	  if (pops[base + chromo].fitness >
	      pops[base + bestinpop[pop]].fitness)
	    bestinpop[pop] = chromo;
	}
    }

}


void permute_slots_reinit (pops, base, range, rrate)
     CHROMOSOME *pops;
     int base, range;
     double rrate;
{
  int chromo,pop, noreinit = 0;
  CHROMOSOME child, old;

  pop = (int) (base / range);
  total_abs_fitness[pop] = 0.0;

  for (chromo = 1; chromo < range; chromo++)
    {
      if (!(pops[base].fitness == pops[base + chromo].fitness))
	{
	  noreinit = 1;
	  break;
	}
    }

  if (noreinit)
    return;			/* If there are some differences, don't reinit */


  if (pops[base].fitness > 0.9)	/* if no differences, but all perfect, */
    return;			/* don't reinit */
  reinits++;
  /* Then reinitialise the population by saving the best, and making the */
  /* rest rrate% mutants of the best */

  for (chromo = base + 1; chromo < base + range; chromo++)
    {
      old = pops[chromo];
      child = permute_slots (pops[chromo]);
      child.fitness = evaluate (child.gdp->genotype,
				old.gdp->genotype,
				&child,
				&old,
				base / numchromo);

      freechromo (&pops[chromo]);
      pops[chromo] = child;
    }


  if (!spatialchosen)
    {
      qsort ((char *) ((int) pops + (base * sizeof (CHROMOSOME))),
	     range, sizeof (CHROMOSOME), compchromo);
      bestinpop[pop] = 0;
    }
  else
    {
      bestinpop[pop] = 0;
      for (chromo = 1; chromo < numchromo; chromo++)
	{
	  if (pops[base + chromo].fitness >
	      pops[base + bestinpop[pop]].fitness)
	    bestinpop[pop] = chromo;
	}
    }

  /*  hctrials = 10000; hillclimb(pops,base,1); hctrials=0; */
}

/***************************************************************************/
/* Reinitialise a population when all genomes have the same fitness.       */
/**************************************************************************/
void converge_reinit (pops, base, range, rrate)
     CHROMOSOME *pops;
     int base, range;
     double rrate;
{
  int chromo, pop, noreinit = 0;
  CHROMOSOME child, old;

  pop = (int) (base / range);
  total_abs_fitness[pop] = 0.0;

  for (chromo = 1; chromo < range; chromo++)
    {
      if (!(pops[base].fitness == pops[base + chromo].fitness))
	{
	  noreinit = 1;
	  break;
	}
    }

  if (noreinit)
    return;			/* If there are some differences, don't reinit */


  if (pops[base].fitness > 0.9)	/* if no differences, but all perfect, */
    return;			/* don't reinit */

  reinits++;
  /* Then reinitialise the population by saving the best, and making the */
  /* rest rrate% mutants of the best */

  for (chromo = base + 1; chromo < base + range; chromo++)
    {
      old = pops[chromo];
      child = mutate (pops[chromo], rrate);
      child.fitness = evaluate (child.gdp->genotype,
				old.gdp->genotype,
				&child,
				&old,
				base / numchromo);

      freechromo (&pops[chromo]);
      pops[chromo] = child;
    }


  if (!spatialchosen)
    {
      qsort ((char *) ((int) pops + (base * sizeof (CHROMOSOME))),
	     range, sizeof (CHROMOSOME), compchromo);
      bestinpop[pop] = 0;
    }
  else
    {
      bestinpop[pop] = 0;
      for (chromo = 1; chromo < numchromo; chromo++)
	{
	  if (pops[base + chromo].fitness >
	      pops[base + bestinpop[pop]].fitness)
	    bestinpop[pop] = chromo;
	}
    }
}






