/************************************************************************
 *                                                                      *
 *               GATT - Genetic Algorithm Time-Tabler                   *
 *                                                                      *
 *                By Dave Corne                                         *
 *                   Peter Ross (peter@aisb.ed.ac.uk),                  *
 *                   Emma Collingwood (emmac@aisb.ed.ac.uk)             *
 *                                                                      *
 ************************************************************************/

/* 
   Copyright (C) 1996, Dave Corne, Peter Ross and Emma Collingwood

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

/**************************************************************************
 * Filename :  reprod.c
 *
 * Reproduction choices. Each takes a CHROMOSOME * pointing to the array
 * of populations, and two ints giving base index and size of target pop
 * in that array. Returns nothing.
 ***************************************************************************/

#define TRUE 1
#define FALSE 0

#include "gatt.h"
extern int simulated_annealing;
extern CHROMOSOME (*selection) ();
extern CHROMOSOME (*crossover) ();
extern CHROMOSOME copychromo ();
extern void copy_chromo ();
extern void create_chromo ();
extern double (*mrate) ();
extern double (*evaluate) ();
extern double (*init_evaluate) ();
extern double best_fitness;
extern double drandom ();
extern double temperature;
extern double insert_prob;
extern void adjust_temperature ();
extern int compchromo ();
extern int numpop;
extern int geno_size;
extern int bestchrom;
extern int stochastic_reproduction;
extern CHROMOSOME (*mutate) ();
extern CHROMOSOME dsm_mutate ();
extern CHROMOSOME sa_mutate ();
extern CHROMOSOME permute_slots ();
extern CHROMOSOME period_swap_mutate ();
extern void insert ();
extern void freechromo ();
extern void update_bests ();
extern FILE *output;
extern FILE *crosspoints;
extern double mute_rate;
extern double cross_rate;
extern double *total_abs_fitness;
extern double reinit_rate;
extern double dsm_rate;
extern double mutation_rate;
extern double perm_rate;
extern int numchromo;
extern int dsm;
extern int ss_xmax, ss_ymax, ss_n;
extern int replace_equal;

extern char *malloc ();
extern double fabs ();
extern int *evals;
extern int *checks;
extern int *bestinpop;
int totalchecks = 0;
int totalevals = 0;
double meanchecks = 0;
double sdchecks = 0;
double meanevals = 0;
double sdevals = 0;
void printstats ();
void recordpoint ();
int reporton ();
extern double *dchecks;
extern double *devals;
extern int trials;
extern void timestart ();
extern void timestop ();
void update_operator_rates ();
int efms;
int sefms;
int crosses;
int scrosses;
int bestpen = 10000000;
int bestclash = 10000000;
int besttc = 10000000;
extern int hctrials;
extern int clashpen;
extern int tooclosepen;
extern int modexcslotpen;
extern int ssloadpen;
OP_RESULT operate ();
extern double sabestfit;
extern int sabest[200];

/* function definitions */
int spatial_stochastic_replace ();
int boltzmann ();

void 
sefm_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  CHROMOSOME parent, child;
  int pop;
  double cf, pf;

  pop = (int) (base / range);
  parent = pops[1];

  /* Perform the mutation, generating the child */
  child = dsm_mutate (parent.gdp->genotype, &parent, base / numchromo);

  pf = parent.fitness;
  cf = child.fitness;
  if (cf > pops[0].fitness)
    {				/* Save this as best so far */
      freechromo (&pops[0]);
      freechromo (&pops[1]);
      pops[0] = child;
      pops[1] = child;
    }
  else if (cf > pf)
    {
      freechromo (&pops[1]);
      pops[1] = child;
    }
  else
    /* Maybe keep it, maybe not; prob based on fitness */
    {

    }				/* Fiddly other bits like reporton, etc ... also deal with initting */
}


void 
one_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  CHROMOSOME child;
  OP_RESULT result;
  int place, ncango, cangos[1000];
  int i, newbest = 0;

  result = operate (pops, base, range, 0, 0, 0);
  child = result.child;
  newbest = result.newbest;


  if (((child.clashes) * clashpen +
       (child.toocloses) * tooclosepen +
       (child.modexcslots) * modexcslotpen) < bestpen)
    {
      bestpen = (child.clashes * clashpen) +
	(child.toocloses) * tooclosepen +
	(child.modexcslots) * modexcslotpen;
      bestclash = child.clashes;
      besttc = child.toocloses;
    }

  /* Choose a random non-best chromosome, and replace that */

  /* Find places it can go */

  ncango = 0;

  for (i = 0; i < numchromo; i++)
    if (pops[i].fitness < pops[bestchrom].fitness)
      cangos[ncango++] = i;

  if (ncango < numchromo / 2)
    place = random ()% range;
  else
    place = cangos[random ()% ncango];

  freechromo (&pops[place]);
  pops[place] = child;

}


/************************************************************************/
/* This function performs elitist generation based repduction. A new    */
/* population is built by selecting members of the old one, possibly    */
/* crossing them with another, and possibly mutating them. This new     */
/* population is sorted and then used to replace the old population.    */
/************************************************************************/

void 
gen_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  CHROMOSOME *newpop, child;
  int loop, pop;
  OP_RESULT result;

  pop = base / numchromo;	/* which population we're reconstructing */

  newpop = (CHROMOSOME *) malloc (sizeof (CHROMOSOME) * (range - 1));

  for (loop = 0; loop < (range - 1);)
    {

      result = operate (pops, base, range, 0, 0, 0);
      child = result.child;
      newpop[loop++] = child;
    }

  /* By this point, newpop is the new pop */

  total_abs_fitness[pop] = 0.0;
  for (loop = 1; loop < range; loop += 1)
    {
      freechromo (&pops[base + loop]);
      pops[base + loop] = newpop[loop - 1];
      total_abs_fitness[pop] += fabs (newpop[loop].fitness);
    }
  free (newpop);
  qsort ((char *) ((int) pops + (base * sizeof (CHROMOSOME))),
	 range, sizeof (CHROMOSOME), compchromo);
}

/* This does `spatial selection' reproduction, distantly based on Wright's
 * shifting-balance model of evolution. The much more common panmictic model
 * is Fisher's in which the entire population competes in selection - eg
 * in roulette-wheel or rank-based selection, the fitness of all members
 * of the population is implicitly used in each selection. In Wright's
 * model, selection need only consider chromosomes in some (2-D) neighbourhood
 * each time. It seems to make sense to use a 2-D grid rather than 3-D or
 * higher; given sufficiently high dimension, every point would be an
 * immediate neighbour of every other so it would be a form of panmictic
 * evolution again.
 *
 * In this implementation, the chromosomes are regarded as being arranged
 * in a 2-D toroidal grid; so first the population size is first automatically
 * adjusted so that the grid can be rectangular and roughly square.
 *
 * Reproduction is generational and proceeds as follows. Each chromosome
 * is replaced by the result of crossover and mutation applied to two parents
 * chosen as the fittest met on a random walk of length ss_n starting from
 * the place where the child will be installed. If you use the -t (twins)
 * flag location (x,y) and location (x,y+1) (if it exists) get used to
 * store the children.
 *
 * This spatial selection reproduction utilises specific selection and
 * insertion algorithms, so you cannot use the -s option with it.
 */


/*******************************************************************/
/* This chooses a location in the neighbourhood of given x,y using */
/* a random walk of length ss_n. Choose the fittest found, ties    */
/* are broken by choosing last.                                    */
/*******************************************************************/

#define LOCATION(x,y) (base+(y)*ss_xmax+(x))

/* The helps cut down on the number of drandom() calls. */
/* Returns random integer in range [0,7].               */
int rand_dir_bits = 0;
int rand_dir_store = 0;
int 
rand_dir ()
{
  int direction;

  if (rand_dir_bits == 0)
    {
      rand_dir_store = (int) (drandom ()* 16777216.0);
      rand_dir_bits = 24;
    }
  direction = rand_dir_store & 0x07;
  rand_dir_store = rand_dir_store >> 3;
  rand_dir_bits -= 3;
  return (direction);
}

CHROMOSOME 
ss_select (pops, base, x, y)
     CHROMOSOME *pops;
     int base, x, y;
{
  int i;
  CHROMOSOME chosen;
  double f;

  chosen = pops[LOCATION (x, y)];
  f = chosen.fitness;
  for (i = 0; i < ss_n; i++)
    {
      switch (rand_dir ())
	{
	case 0:
	  y -= 1;
	  break;
	case 1:
	  x += 1;
	  y -= 1;
	  break;
	case 2:
	  x += 1;
	  break;
	case 3:
	  x += 1;
	  y += 1;
	  break;
	case 4:
	  y += 1;
	  break;
	case 5:
	  x -= 1;
	  y += 1;
	  break;
	case 6:
	  x -= 1;
	  break;
	case 7:
	  x -= 1;
	  y -= 1;
	  break;
	}

      /* Note that it is probably cheaper to do the next two lines */
      /* than to use conditional tests of whether x or y is out of */
      /* bounds.                                                   */
      x = (x + ss_xmax) % ss_xmax;
      y = (y + ss_ymax) % ss_ymax;

      if (pops[LOCATION (x, y)].fitness >= f)
	{
	  chosen = pops[LOCATION (x, y)];
	  f = chosen.fitness;
	}
    }
  return (chosen);
}


void 
evolutionary_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  int i, j;
  CHROMOSOME parent, child;
  CHROMOSOME *interimpop;

  /* Somewhere to keep the interim population */

  interimpop = (CHROMOSOME *) malloc (range * (trials + 1) * sizeof (CHROMOSOME));

  /* Put <trials> mutants of each member of the pop into the interimpop */

  for (i = 0; i < range; i++)
    {
      parent = pops[base + i];
      for (j = 0; j < trials; j++)
	{
	  child = dsm_mutate (parent.gdp->genotype,
			      &parent, base / range);
	  reporton (pops, child.fitness, (int) (base / numchromo));
	  interimpop[range * j + i] = child;
	}
    }

  /* Put a copy of each parent into the interimpop */
  for (i = 0; i < range; i++)
    interimpop[trials * range + i] = copychromo (pops[base + i]);


  /* Sort the interimpop */
  qsort ((char *) (int) interimpop, range * (trials + 1),
	 sizeof (CHROMOSOME), compchromo);

  /* make sure the best is retained */
  freechromo (&pops[base]);
  pops[base] = copychromo (interimpop[0]);

  /* Now do selection on interimpop to fill up pop */

  for (i = 1; i < range; i++)
    {
      freechromo (&pops[base + i]);
      child = selection (interimpop, 0, (trials + 1) * range);
      pops[base + i] = copychromo (child);
    }
  /* Now free interimpop */

  for (i = 0; i < (trials + 1) * range; i++)
    freechromo (&interimpop[i]);

  free (interimpop);

}


/*******************************************************************/
/* Spatial selection. Uses specific selection and insertion, so    */
/* you cannot set them independently. Generational version.        */
/*******************************************************************/

void 
ss_gen_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  int x, y;
  CHROMOSOME parent1, parent2, child;
  CHROMOSOME *newpop;

  /* Somewhere to keep the new population, so that all selection is */
  /* from the old population.                                       */
  newpop = (CHROMOSOME *) malloc (range * sizeof (CHROMOSOME));

  /* Scan the grid, doing selection and insertion */
  for (x = 0; x < ss_xmax; x++)
    {
      for (y = 0; y < ss_ymax;)
	{
	  parent1 = ss_select (pops, base, x, y);
	  parent2 = ss_select (pops, base, x, y);
	  child = crossover (parent1, parent2);
	  child = mutate (child, mrate (parent1, parent2));
	  child.fitness = evaluate (child.gdp->genotype,
				    parent1.gdp->genotype,
				    &child,
				    &parent1,
				    base/numchromo);
	  child.gdp->refcount++;
	  newpop[y * ss_xmax + x] = child;
	  y++;

	}
    }



  /* Copy new that is better into old, freeing any unwanted old stuff. */
  for (x = 0; x < numchromo; x++)
    {
      if (pops[base + x].fitness < newpop[x].fitness)
	{
	  if (pops[base + x].gdp->refcount == 1)
	    {
	      free (pops[base + x].gdp->genotype);
	      free (pops[base + x].gdp);
	    }
	  else
	    pops[base + x].gdp->refcount--;
	  pops[base + x] = newpop[x];
	}
      else
	{
	  free (newpop[x].gdp->genotype);
	  free (newpop[x].gdp);
	}
    }

  free (newpop);
}


/*******************************************************************/
/* Spatial selection. Uses specific selection and insertion, so    */
/* you cannot set them independently. One (Genitor-ish) version.   */
/*******************************************************************/

void 
mix_ops_ss_one_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  int pop, x, y, newbest = 0, last = 0, chk;
  double r;
  CHROMOSOME parent1, parent2, child;

  pop = (int) base / numchromo;
  chk = checks[pop];
  x = drandom ()* (double) ss_xmax;
  y = drandom ()* (double) ss_ymax;

  parent1 = ss_select (pops, base, x, y);

  r = drandom ();

  if (r <= dsm_rate)
    {
      child = dsm_mutate (parent1.gdp->genotype,
			  &parent1, base / numchromo);
      efms++;
      if (child.fitness > parent1.fitness)
	sefms += (int) ((int) (1.0 / parent1.fitness) - (int) (1.0 / child.fitness));
      last = 0;
    }
  else if (r <= (dsm_rate + mutation_rate))
    {
      child = mutate (parent1, mute_rate);
      child.fitness = evaluate (child.gdp->genotype,
				parent1.gdp->genotype,
				&child,
				&parent1,
				base / numchromo);
    }
  else
    {
      parent2 = ss_select (pops, base, x, y);
      child = crossover (parent1, parent2);
      child.fitness = evaluate (child.gdp->genotype,
				parent1.gdp->genotype,
				&child,
				&parent1,
				base / numchromo);

      crosses++;
      if ((child.fitness > parent1.fitness) &&
	  (child.fitness > parent2.fitness))
	scrosses += (int) ((int) (1.0 / parent1.fitness) - (int) (1.0 / child.fitness));
      last = 1;
    }
  newbest = reporton (pops, child.fitness, (int) (base / numchromo));

  if (newbest)
    bestinpop[(int) (base / range)] = LOCATION (x, y) - base;

  if (newbest || (!stochastic_reproduction))
    {
      if (pops[LOCATION (x, y)].fitness <= child.fitness)
	{
	  freechromo (&pops[LOCATION (x, y)]);
	  pops[LOCATION (x, y)] = child;
	}
      else
	freechromo (&child);
    }
  else
    {
      if (spatial_stochastic_replace (pops, base, range, child, x, y))
	{
	  freechromo (&pops[LOCATION (x, y)]);
	  pops[LOCATION (x, y)] = child;
	}
      else
	freechromo (&child);
    }
  update_operator_rates ((int) base / numchromo, last, (checks[pop] - chk));
}


void 
crossexpts (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  int x, y, newbest = 0;
  CHROMOSOME parent1, parent2, child;

  for (x = 0; x < numchromo; x++)
    {
      parent1 = pops[base + x];
      child = dsm_mutate (parent1.gdp->genotype,
			  &parent1, base / numchromo);
      freechromo (&parent1);
      pops[base + x] = child;
      newbest = reporton (pops, child.fitness, (int) (base / numchromo));
      if (newbest)
	bestinpop[(int) (base / range)] = x;
    }

  for (x = 0; x < numchromo - 1; x++)
    for (y = x + 1; y < numchromo; y++)
      {
	parent1 = pops[base + x];
	parent2 = pops[base + y];
	child = crossover (parent1, parent2);
	child.fitness = evaluate (child.gdp->genotype,
				  parent1.gdp->genotype,
				  &child,
				  &parent1,
				  base / numchromo);
	recordpoint (parent1, parent2, child);
      }
}

void hillclimb (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  int x, trial, newbest = 0;
  CHROMOSOME parent, child;

  for (x = base; x < base + range; x++)
    {
      for (trial = 0; trial < hctrials; trial++)
	{
	  parent = pops[x];
	  child = dsm_mutate (parent.gdp->genotype,
			      &parent, base / numchromo);
	  if (child.fitness > parent.fitness)
	    {
	      freechromo (&parent);
	      pops[x] = child;
	      newbest = reporton (pops, child.fitness, (int) (base / numchromo));
	      if (newbest)
		bestinpop[(int) (base / range)] = x - base;
	    }
	}
    }
}

void 
ss_one_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  int k, x, y, oldwasbest = 0, newbest = 0;
  double cp, pp;
  CHROMOSOME child;
  OP_RESULT result;


  x = drandom ()* (double) ss_xmax;
  y = drandom ()* (double) ss_ymax;

  if (pops[LOCATION (x, y)].fitness == best_fitness)
    oldwasbest = 1;

  /*   parent1 = ss_select(pops,base,x,y); */

  result = operate (pops, base, range, 1, x, y);
  child = result.child;
  newbest = result.newbest;


  if (newbest)
    bestinpop[(int) (base / range)] = LOCATION (x, y) - base;

  if (((child.clashes) * clashpen +
       (child.toocloses) * tooclosepen +
       (child.modexcslots) * modexcslotpen +
       (child.ssload) * ssloadpen) < bestpen)
    {
      bestpen = (child.clashes * clashpen) +
	(child.toocloses) * tooclosepen +
	(child.modexcslots) * modexcslotpen +
	(child.ssload) * ssloadpen;
      bestclash = child.clashes;
      besttc = child.toocloses;
    }



  /* If this child is fitter or equal to the current (x,y) resident */
  /* then insert it. Else, with probability insert_prob, insert it */
  /* by replacing the least fit of the located chromo, or a neighbour */
  /* making sure that the bestinpop is dealt with properly */

  if ((child.fitness > pops[LOCATION (x, y)].fitness) ||
      ((replace_equal == 1) &&
       (child.fitness == pops[LOCATION (x, y)].fitness)))
    {
      if (simulated_annealing)
	if (child.fitness > sabestfit)
	  {
	    for (k = 0; k < nummods; k++)
	      sabest[k] = child.gdp->genotype[k];
	    sabestfit = child.fitness;
	  }
      freechromo (&pops[LOCATION (x, y)]);
      pops[LOCATION (x, y)] = child;
    }
  else
    {				/* Child is less fit than location, but maybe insert */
      /* anyway; do so if current location is not bestinpop */
      /* and with probability insert_prob */
      pp = (1.0 / pops[LOCATION (x, y)].fitness - 1);
      cp = (1.0 / child.fitness - 1);
      if ((((LOCATION (x, y) - base) != bestinpop[(int) base / range]) &&
	   (drandom ()<= insert_prob))
	  ||
	  ((numchromo == 1) && (simulated_annealing == 1) &&
	   (drandom ()< exp ((pp - cp) / temperature))))
	{
	  freechromo (&pops[LOCATION (x, y)]);
	  pops[LOCATION (x, y)] = child;
	}
      else
	freechromo (&child);
    }

  if (simulated_annealing)
    adjust_temperature();
}

void 
ss_one_boltzmann_reproduction (pops, base, range)
     CHROMOSOME *pops;
     int base, range;
{
  int x, y, newbest = 0;
  CHROMOSOME child, *tmp;
  OP_RESULT result;

  x = drandom ()* (double) ss_xmax;
  y = drandom ()* (double) ss_ymax;

  adjust_temperature ();

  if (simulated_annealing)
    result = operate (pops, base, range, 2, x, y);
  else
    result = operate (pops, base, range, 1, x, y);

  child = result.child;
  newbest = result.newbest;

  if (simulated_annealing)
    {
      if (newbest)
	{
	  freechromo (&pops[0]);
	  pops[0] = child;
	  freechromo (&pops[1]);

	  tmp = (CHROMOSOME *) malloc (sizeof (CHROMOSOME));
	  create_chromo (tmp);
	  copy_chromo (tmp, &child);
	  pops[1] = *tmp;
	  return;
	}
      else if ((child.fitness == pops[1].fitness) ||
	       (boltzmann ((1.0 / child.fitness), (1.0 / pops[1].fitness))))
	{
	  freechromo (&pops[1]);
	  pops[1] = child;
	  return;
	}
      else
	{
	  freechromo (&child);
	  return;
	}
    }

  if (newbest)
    bestinpop[(int) (base / range)] = LOCATION (x, y) - base;

  /* If this child is fitter or equal to the current (x,y) resident */
  /* then insert it. */

  if ((child.fitness > pops[LOCATION (x, y)].fitness) ||
      ((replace_equal == 1) &&
       (child.fitness == pops[LOCATION (x, y)].fitness)))
    {
      freechromo (&pops[LOCATION (x, y)]);
      pops[LOCATION (x, y)] = child;
    }
  else
    /* Have a Boltzmann tournament between this child and */
    /* the resident, unless the resident doesn't happen to */
    /* be the nominated best chrom */ if ((LOCATION (x, y) != bestchrom) &&
  (boltzmann ((1.0 / child.fitness), (1.0 / pops[LOCATION (x, y)].fitness))))
    {
      freechromo (&pops[LOCATION (x, y)]);
      pops[LOCATION (x, y)] = child;
    }
  else
    freechromo (&child);
}

int 
boltzmann (double newcost, double oldcost)
{
  if (drandom ()< exp ((oldcost - newcost) / temperature))
    return 1;
  else
    return 0;
}

int 
spatial_stochastic_replace (pops, base, range, child, x, y)
     CHROMOSOME *pops;
     CHROMOSOME child;
     int base, range, x, y;
{
  double fchild, prob = 0.01;

  fchild = child.fitness;
  if (pops[LOCATION (x, y)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x + 1, y)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x - 1, y)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x, y + 1)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x + 1, y + 1)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x - 1, y + 1)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x, y - 1)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x + 1, y - 1)].fitness < fchild)
    prob += 0.1;
  if (pops[LOCATION (x - 1, y - 1)].fitness < fchild)
    prob += 0.1;

  if (drandom ()<= prob)
    return 1;
  else
    return 0;
}


int 
reporton (CHROMOSOME * pops, double fitness, int pop)
{


  if ((fitness > 0.9) &&
      (checks[pop] > 0))	/* A hack */
    {
      if (output)
	fprintf (output, "Opt found in pop %d in %d evals %d checks\n",
		 pop, evals[pop], checks[pop]);
      dchecks[pop] = 2 * (double) checks[pop];
      devals[pop] = (double) evals[pop];
      checks[pop] = -1000000000;
    }

  if (fitness > pops[pop * numchromo + bestinpop[pop]].fitness)
    return 1;
  else
    return 0;

}

void 
printstats ()
{
  int i;
  double checkvar, evalvar;

  checkvar = evalvar = 0.0;

  for (i = 0; i < numpop; i++)
    {
      totalchecks += dchecks[i];
      totalevals += devals[i];
    }
  meanchecks = (double) totalchecks / numpop;
  meanevals = (double) totalevals / numpop;

  for (i = 0; i < numpop; i++)
    {
      dchecks[i] -= meanchecks;
      dchecks[i] *= dchecks[i];
      checkvar += dchecks[i];

      devals[i] -= meanevals;
      devals[i] *= devals[i];
      evalvar += devals[i];
    }

  evalvar /= numpop;
  checkvar /= numpop;

  fprintf (output, "Mean evals %g ; sd %g \n Mean checks %g ; sd %g\n",
	   meanevals, sqrt (evalvar), meanchecks, sqrt (checkvar));
}

void 
update_operator_rates (int pop, int last, int chk)
{
  double efmsr, crosssr, avetime, avctime;
  static int etime = 0;
  static int ctime = 0;

  if (last)
    ctime += chk + 10;
  else
    etime += chk + 10;


  if ((efms + crosses) >= 500)
    {
      avetime = etime / (double) efms;
      avctime = ctime / (double) crosses;
      etime = ctime = 0;

      efmsr = (double) ((double) sefms / (double) efms);
      crosssr = (double) ((double) scrosses / (double) crosses);

      efmsr /= avetime;
      crosssr /= avctime;
      efms = crosses = 0;
      sefms = scrosses = 0;

      if (efmsr >= crosssr)
	dsm_rate = 0.7;
      else
	dsm_rate = 0.1;
      fprintf (output, " %d d %g esr%g csr%g ave%g avc%g\n", evals[pop], dsm_rate,
	       efmsr, crosssr, avetime, avctime);

    }

}


void 
recordpoint (p1, p2, c, f)
     CHROMOSOME p1, p2, c;
     FILE *f;
{
  double p1f, p2f, apf, cf;

  p1f = (1.0 / (p1.fitness) - 1.0);
  p2f = (1.0 / (p2.fitness) - 1.0);
  cf = (1.0 / (c.fitness) - 1.0);

  /*   apf = (p1f+p2f)/2.0;                 apf is average of parents  */

  if (p1f <= p2f)
    apf = p1f;
  else
    apf = p2f;			/* apf is fittest parent */
  fprintf (f, "%g %g \n", apf, cf);
}

OP_RESULT 
operate (pops, base, range, spatial, x, y)
     CHROMOSOME *pops;
     int spatial, base, range, x, y;
{

  double r;
  CHROMOSOME parent1, parent2, child;
  OP_RESULT result;
  int newbest;

  if (spatial == 0)
    parent1 = selection (pops, base, range);	/* Select a parent normally */
  else if (spatial == 1)
    parent1 = ss_select (pops, base, x, y);	/* Select a parent spatially */
  else if (spatial == 2)
    parent1 = pops[1];		/* Simulated annealing */

  r = drandom ();		/* Generate a random number in [0,1]*/

  if (spatial == 2)
    {
      child = sa_mutate (parent1.gdp->genotype,
			 &parent1, base / numchromo);
    }
  else if (r <= dsm_rate)
    {
      child = dsm_mutate (parent1.gdp->genotype,
			  &parent1, base / numchromo);
    }
  else if (r <= (dsm_rate + mutation_rate))
    {
      child = period_swap_mutate (parent1.gdp->genotype, parent1, 0, 1);
      child.fitness = evaluate (child.gdp->genotype,
				parent1.gdp->genotype,
				&child,
				&parent1,
				base / numchromo);
    }
  else if (r <= (dsm_rate + mutation_rate + perm_rate))
    {
      child = permute_slots (parent1);
      child.fitness = evaluate (child.gdp->genotype,
				parent1.gdp->genotype,
				&child,
				&parent1,
				base / numchromo);
    }
  else
    {

      if (spatial == 0)
	parent2 = selection (pops, base, range);	/* Select a parent normally */
      else
	parent2 = ss_select (pops, base, x, y);	/* Select a parent spatially */

      child = crossover (parent1, parent2);	/* Cross them. */

      child.fitness = evaluate (child.gdp->genotype,
				parent1.gdp->genotype,
				&child,
				&parent1,
				base / numchromo);
    }

  newbest = reporton (pops, child.fitness, (int) (base / numchromo));

  result.child = child;
  result.newbest = newbest;

  return (result);
}
